@@ -2,22 +2,25 @@
  *      menu-cached.c
  *
  *      Copyright 2008 - 2010 PCMan <pcman.tw@gmail.com>
- *      Copyright 2012-2013 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+ *      Copyright 2009 Jürgen Hötzel <juergen@archlinux.org>
+ *      Copyright 2012-2015 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
- *      This program is free software; you can redistribute it and/or modify
- *      it under the terms of the GNU General Public License as published by
- *      the Free Software Foundation; either version 2 of the License, or
- *      (at your option) any later version.
+ *      This file is a part of libmenu-cache package and created program
+ *      should be not used without the library.
  *
- *      This program is distributed in the hope that it will be useful,
+ *      This library is free software; you can redistribute it and/or
+ *      modify it under the terms of the GNU Lesser General Public
+ *      License as published by the Free Software Foundation; either
+ *      version 2.1 of the License, or (at your option) any later version.
+ *
+ *      This library is distributed in the hope that it will be useful,
  *      but WITHOUT ANY WARRANTY; without even the implied warranty of
- *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *      GNU General Public License for more details.
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *      Lesser General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      You should have received a copy of the GNU Lesser General Public
+ *      License along with this library; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -140,7 +143,8 @@ static gboolean read_all_used_files( FILE* f, int* n_files, char*** used_files )
         return FALSE;
     if( sscanf(line, "%d.%d", &ver_maj, &ver_min)< 2 )
         return FALSE;
-    if( ver_maj != VER_MAJOR || ver_min != VER_MINOR )
+    if (ver_maj != VER_MAJOR ||
+        ver_min > VER_MINOR || ver_min < VER_MINOR_SUPPORTED)
         return FALSE;
 
     /* skip the second line containing menu name */
@@ -157,7 +161,7 @@ static gboolean read_all_used_files( FILE* f, int* n_files, char*** used_files )
     for( i = 0, x = 0; i < n; ++i )
     {
         int len;
-        GFile *gfile;
+
         if( ! fgets( line, G_N_ELEMENTS(line), f ) )
             return FALSE;
 
@@ -165,8 +169,7 @@ static gboolean read_all_used_files( FILE* f, int* n_files, char*** used_files )
         if( len <= 1 )
             return FALSE;
         files[ x ] = g_strndup( line, len - 1 ); /* don't include \n */
-        gfile = g_file_new_for_path(files[x]+1);
-        if (g_file_query_exists(gfile, NULL))
+        if (g_file_test(files[x]+1, G_FILE_TEST_EXISTS))
             x++;
         else
         {
@@ -174,7 +177,6 @@ static gboolean read_all_used_files( FILE* f, int* n_files, char*** used_files )
             g_free(files[x]);
             files[x] = NULL;
         }
-        g_object_unref(gfile);
     }
     *n_files = x;
     *used_files = files;
@@ -393,9 +395,11 @@ static gboolean cache_file_is_updated( const char* cache_file, int* n_used_files
 {
     gboolean ret = FALSE;
     struct stat st;
+#if 0
     time_t cache_mtime;
     char** files;
     int n, i;
+#endif
     FILE* f;
 
     f = fopen( cache_file, "r" );
@@ -403,6 +407,7 @@ static gboolean cache_file_is_updated( const char* cache_file, int* n_used_files
     {
         if( fstat( fileno(f), &st) == 0 )
         {
+#if 0
             cache_mtime = st.st_mtime;
             if( read_all_used_files(f, &n, &files) )
             {
@@ -421,6 +426,9 @@ static gboolean cache_file_is_updated( const char* cache_file, int* n_used_files
                     *used_files = files;
                 }
             }
+#else
+            ret = read_all_used_files(f, n_used_files, used_files);
+#endif
         }
         fclose( f );
     }
@@ -433,7 +441,7 @@ static void get_socket_name( char* buf, int len )
     if(dpy && *dpy)
     {
         char* p = strchr(dpy, ':');
-        for(++p; *p && *p != '.';)
+        for(++p; *p && *p != '.' && *p != '\n';)
             ++p;
         if(*p)
             *p = '\0';
@@ -589,6 +597,13 @@ retry:
                     DEBUG("regeneration of cache failed!!");
                 }
             }
+            else
+            {
+                /* file loaded, schedule update anyway */
+                cache->need_reload = TRUE;
+                cache->delayed_reload_handler = g_timeout_add_seconds_full(G_PRIORITY_LOW, 3,
+                                                    (GSourceFunc)delayed_reload, cache, NULL);
+            }
             memcpy( cache->md5, md5, 33 );
             cache->n_files = n_files;
             cache->files = files;
