@@ -1,7 +1,7 @@
 /*
  *      menu-compose.c : scans appropriate .desktop files and composes cache file.
  *
- *      Copyright 2014 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
+ *      Copyright 2014-2015 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
  *      This file is a part of libmenu-cache package and created program
  *      should be not used without the library.
@@ -26,6 +26,7 @@
 #endif
 
 #include "menu-tags.h"
+#include "version.h"
 
 #include <string.h>
 #include <stdio.h>
@@ -35,6 +36,8 @@
 
 static GSList *DEs = NULL;
 
+static guint req_version = 1; /* old compatibility default */
+
 static void menu_app_reset(MenuApp *app)
 {
     g_free(app->filename);
@@ -45,8 +48,10 @@ static void menu_app_reset(MenuApp *app)
     g_free(app->icon);
     g_free(app->generic_name);
     g_free(app->exec);
+    g_free(app->try_exec);
     g_free(app->wd);
     g_free(app->categories);
+    g_free(app->keywords);
     g_free(app->show_in);
     g_free(app->hide_in);
 }
@@ -62,6 +67,33 @@ static void menu_app_free(gpointer data)
     g_slice_free(MenuApp, app);
 }
 
+static char *_escape_lf(char *str)
+{
+    char *c;
+
+    if (str != NULL && (c = strchr(str, '\n')) != NULL)
+    {
+        GString *s = g_string_new_len(str, c - str);
+
+        while (*c)
+        {
+            if (*c == '\n')
+                g_string_append(s, "\\n");
+            else
+                g_string_append_c(s, *c);
+            c++;
+        }
+        g_free(str);
+        str = g_string_free(s, FALSE);
+    }
+    return str;
+}
+
+static char *_get_string(GKeyFile *kf, const char *key)
+{
+    return _escape_lf(g_key_file_get_string(kf, G_KEY_FILE_DESKTOP_GROUP, key, NULL));
+}
+
 /* g_key_file_get_locale_string is too much limited so implement replacement */
 static char *_get_language_string(GKeyFile *kf, const char *key)
 {
@@ -71,12 +103,45 @@ static char *_get_language_string(GKeyFile *kf, const char *key)
     for (lang = languages; lang[0] != NULL; lang++)
     {
         try_key = g_strdup_printf("%s[%s]", key, lang[0]);
-        str = g_key_file_get_string(kf, G_KEY_FILE_DESKTOP_GROUP, try_key, NULL);
+        str = _get_string(kf, try_key);
         g_free(try_key);
         if (str != NULL)
             return str;
     }
-    return g_key_file_get_locale_string(kf, G_KEY_FILE_DESKTOP_GROUP, key, languages[0], NULL);
+    return _escape_lf(g_key_file_get_locale_string(kf, G_KEY_FILE_DESKTOP_GROUP,
+                                                   key, languages[0], NULL));
+}
+
+static char **_get_string_list(GKeyFile *kf, const char *key, gsize *lp)
+{
+    char **str, **p;
+
+    str = g_key_file_get_string_list(kf, G_KEY_FILE_DESKTOP_GROUP, key, lp, NULL);
+    if (str != NULL)
+        for (p = str; p[0] != NULL; p++)
+            p[0] = _escape_lf(p[0]);
+    return str;
+}
+
+static char **_get_language_string_list(GKeyFile *kf, const char *key, gsize *lp)
+{
+    char **lang;
+    char *try_key, **str;
+
+    for (lang = languages; lang[0] != NULL; lang++)
+    {
+        try_key = g_strdup_printf("%s[%s]", key, lang[0]);
+        str = _get_string_list(kf, try_key, lp);
+        g_free(try_key);
+        if (str != NULL)
+            return str;
+    }
+    str = g_key_file_get_locale_string_list(kf, G_KEY_FILE_DESKTOP_GROUP, key,
+                                            languages[0], lp, NULL);
+    if (str != NULL)
+        for (lang = str; lang[0] != NULL; lang++)
+            lang[0] = _escape_lf(lang[0]);
+    return str;
 }
 
 static void _fill_menu_from_file(MenuMenu *menu, const char *path)
@@ -90,8 +155,7 @@ static void _fill_menu_from_file(MenuMenu *menu, const char *path)
         goto exit;
     menu->title = _get_language_string(kf, G_KEY_FILE_DESKTOP_KEY_NAME);
     menu->comment = _get_language_string(kf, G_KEY_FILE_DESKTOP_KEY_COMMENT);
-    menu->icon = g_key_file_get_string(kf, G_KEY_FILE_DESKTOP_GROUP,
-                                       G_KEY_FILE_DESKTOP_KEY_ICON, NULL);
+    menu->icon = _get_string(kf, G_KEY_FILE_DESKTOP_KEY_ICON);
     menu->layout.nodisplay = g_key_file_get_boolean(kf, G_KEY_FILE_DESKTOP_GROUP,
                                                     G_KEY_FILE_DESKTOP_KEY_NO_DISPLAY, NULL);
     menu->layout.is_set = TRUE;
@@ -100,12 +164,17 @@ exit:
 }
 
 static const char **menu_app_intern_key_file_list(GKeyFile *kf, const char *key,
+                                                  gboolean localized,
                                                   gboolean add_to_des)
 {
     gsize len, i;
-    char **val = g_key_file_get_string_list(kf, G_KEY_FILE_DESKTOP_GROUP, key, &len, NULL);
+    char **val;
     const char **res;
 
+    if (localized)
+        val = _get_language_string_list(kf, key, &len);
+    else
+        val = _get_string_list(kf, key, &len);
     if (val == NULL)
         return NULL;
     res = (const char **)g_new(char *, len + 1);
@@ -124,15 +193,18 @@ static void _fill_app_from_key_file(MenuApp *app, GKeyFile *kf)
 {
     app->title = _get_language_string(kf, G_KEY_FILE_DESKTOP_KEY_NAME);
     app->comment = _get_language_string(kf, G_KEY_FILE_DESKTOP_KEY_COMMENT);
-    app->icon = g_key_file_get_string(kf, G_KEY_FILE_DESKTOP_GROUP,
-                                      G_KEY_FILE_DESKTOP_KEY_ICON, NULL);
+    app->icon = _get_string(kf, G_KEY_FILE_DESKTOP_KEY_ICON);
     app->generic_name = _get_language_string(kf, G_KEY_FILE_DESKTOP_KEY_GENERIC_NAME);
-    app->exec = g_key_file_get_string(kf, G_KEY_FILE_DESKTOP_GROUP,
-                                      G_KEY_FILE_DESKTOP_KEY_EXEC, NULL);
-    //app->wd =
-    app->categories = menu_app_intern_key_file_list(kf, G_KEY_FILE_DESKTOP_KEY_CATEGORIES, FALSE);
-    app->show_in = menu_app_intern_key_file_list(kf, G_KEY_FILE_DESKTOP_KEY_ONLY_SHOW_IN, TRUE);
-    app->hide_in = menu_app_intern_key_file_list(kf, G_KEY_FILE_DESKTOP_KEY_NOT_SHOW_IN, TRUE);
+    app->exec = _get_string(kf, G_KEY_FILE_DESKTOP_KEY_EXEC);
+    app->try_exec = _get_string(kf, G_KEY_FILE_DESKTOP_KEY_TRY_EXEC);
+    app->wd = _get_string(kf, G_KEY_FILE_DESKTOP_KEY_PATH);
+    app->categories = menu_app_intern_key_file_list(kf, G_KEY_FILE_DESKTOP_KEY_CATEGORIES,
+                                                    FALSE, FALSE);
+    app->keywords = menu_app_intern_key_file_list(kf, "Keywords", TRUE, FALSE);
+    app->show_in = menu_app_intern_key_file_list(kf, G_KEY_FILE_DESKTOP_KEY_ONLY_SHOW_IN,
+                                                 FALSE, TRUE);
+    app->hide_in = menu_app_intern_key_file_list(kf, G_KEY_FILE_DESKTOP_KEY_NOT_SHOW_IN,
+                                                 FALSE, TRUE);
     app->use_terminal = g_key_file_get_boolean(kf, G_KEY_FILE_DESKTOP_GROUP,
                                                G_KEY_FILE_DESKTOP_KEY_TERMINAL, NULL);
     app->use_notification = g_key_file_get_boolean(kf, G_KEY_FILE_DESKTOP_GROUP,
@@ -145,6 +217,21 @@ static GHashTable *all_apps = NULL;
 
 static GSList *loaded_dirs = NULL;
 
+static GList *_make_def_layout(void)
+{
+    MenuMerge *mm;
+    GList *layout;
+
+    mm = g_slice_new(MenuMerge);
+    mm->type = MENU_CACHE_TYPE_NONE;
+    mm->merge_type = MERGE_FILES;
+    layout = g_list_prepend(NULL, mm);
+    mm = g_slice_new(MenuMerge);
+    mm->type = MENU_CACHE_TYPE_NONE;
+    mm->merge_type = MERGE_MENUS;
+    return g_list_prepend(layout, mm);
+}
+
 static void _fill_apps_from_dir(MenuMenu *menu, GList *lptr, GString *prefix,
                                 gboolean is_legacy)
 {
@@ -178,10 +265,8 @@ static void _fill_apps_from_dir(MenuMenu *menu, GList *lptr, GString *prefix,
             if (is_legacy)
             {
                 MenuMenu *submenu = g_slice_new0(MenuMenu);
-                MenuMerge def_files = { .type = MENU_CACHE_TYPE_NONE, .merge_type = MERGE_FILES };
-                MenuMerge def_menus = { .type = MENU_CACHE_TYPE_NONE, .merge_type = MERGE_MENUS };
                 submenu->layout = menu->layout; /* copy all */
-                submenu->layout.items = g_list_prepend(g_list_prepend(NULL, &def_files), &def_menus);
+                submenu->layout.items = _make_def_layout(); /* default layout */
                 submenu->layout.inline_limit_is_set = TRUE; /* marker */
                 submenu->name = g_strdup(name);
                 submenu->dir = g_intern_string(filename);
@@ -636,21 +721,29 @@ static void _stage1(MenuMenu *menu, GList *dirs, GList *apps, GList *legacy, GLi
         case MENU_CACHE_TYPE_APP: /* MemuFilename */
             VDBG("composing Filename %s", ((MenuFilename *)app)->id);
             app = g_hash_table_lookup(all_apps, ((MenuFilename *)app)->id);
-            if (app == NULL || !app->matched || app->excluded)
+            if (app == NULL)
                 /* not available, ignoring it */
                 break;
             l = g_list_find(result, app); /* this might be slow but we have
                                              to do this because app might be
                                              already added into result */
             if (l != NULL)
+            {
                 /* move it out to this place */
                 result = g_list_remove_link(result, l);
+                VVDBG("+++ composing app %s (move)", app->id);
+            }
             else
             {
                 l = g_list_find(available, app);
+                VVDBG("+++ composing app %s%s", app->id, (l == NULL) ? " (add)" : "");
                 if (l != NULL)
                     available = g_list_remove_link(available, l);
+                else
+                    l = g_list_prepend(NULL, app);
             }
+            if (l != NULL)
+                app->menus = g_list_prepend(app->menus, menu);
             result = g_list_concat(l, result);
             break;
         case MENU_CACHE_TYPE_SEP: /* MenuSep */
@@ -660,6 +753,7 @@ static void _stage1(MenuMenu *menu, GList *dirs, GList *apps, GList *legacy, GLi
         case MENU_CACHE_TYPE_NONE: /* MenuMerge */
             VDBG("composing Merge type %d", ((MenuMerge *)app)->merge_type);
             next = NULL;
+            tag = 0;
             switch (((MenuMerge *)app)->merge_type) {
             case MERGE_FILES:
                 tag = 1; /* use it as mark to not add dirs */
@@ -682,7 +776,7 @@ static void _stage1(MenuMenu *menu, GList *dirs, GList *apps, GList *legacy, GLi
                 available = NULL;
                 /* continue with menus */
             case MERGE_MENUS:
-                for (l = menu->children; l; )
+                if (tag != 1) for (l = menu->children; l; )
                 {
                     if (((MenuMenu *)l->data)->layout.type == MENU_CACHE_TYPE_DIR)
                     {
@@ -805,8 +899,7 @@ static gint _stage2(MenuMenu *menu, gboolean with_hidden)
             /* do recursion */
             if (_stage2(child->data, with_hidden) > 0)
                 count++;
-            else
-                /* if (!with_hidden || req_version < 2) */
+            else if (!with_hidden || req_version < 2)
                 to_delete = g_list_prepend(to_delete, child);
             child = next;
             break;
@@ -833,6 +926,8 @@ static gint _stage2(MenuMenu *menu, gboolean with_hidden)
     {
         if (menu->layout.show_empty)
             count++;
+        else
+            menu->layout.nodisplay = TRUE;
     }
     return count;
 }
@@ -850,6 +945,23 @@ static inline int _compose_flags(const char **f)
     return x;
 }
 
+static gboolean write_app_extra(FILE *f, MenuApp *app)
+{
+    gboolean ret;
+    char *cats, *keywords;
+    char *null_list[] = { NULL };
+
+    if (req_version < 2)
+        return TRUE;
+    cats = g_strjoinv(";", app->categories ? (char **)app->categories : null_list);
+    keywords = g_strjoinv(",", app->keywords ? (char **)app->keywords : null_list);
+    ret = fprintf(f, "%s\n%s\n%s\n%s\n", NONULL(app->try_exec), NONULL(app->wd),
+                                         cats, keywords) > 0;
+    g_free(cats);
+    g_free(keywords);
+    return ret;
+}
+
 static gboolean write_app(FILE *f, MenuApp *app, gboolean with_hidden)
 {
     int index;
@@ -872,7 +984,7 @@ static gboolean write_app(FILE *f, MenuApp *app, gboolean with_hidden)
     return fprintf(f, "-%s\n%s\n%s\n%s\n%s\n%d\n%s\n%s\n%u\n%d\n", app->id,
                    NONULL(app->title), NONULL(app->comment), NONULL(app->icon),
                    NONULL(app->filename), index, NONULL(app->generic_name),
-                   NONULL(app->exec), flags, show) > 0;
+                   NONULL(app->exec), flags, show) > 0 && write_app_extra(f, app);
 }
 
 static gboolean write_menu(FILE *f, MenuMenu *menu, gboolean with_hidden)
@@ -883,14 +995,17 @@ static gboolean write_menu(FILE *f, MenuMenu *menu, gboolean with_hidden)
 
     if (!with_hidden && !menu->layout.show_empty && menu->children == NULL)
         return TRUE;
-    if (menu->layout.nodisplay/* && (!with_hidden || req_version < 2)*/)
+    if (menu->layout.nodisplay && (!with_hidden || req_version < 2))
         return TRUE;
     index = g_slist_index(DirDirs, menu->dir);
     if (fprintf(f, "+%s\n%s\n%s\n%s\n%s\n%d\n", menu->name, NONULL(menu->title),
                 NONULL(menu->comment), NONULL(menu->icon),
                 menu->id ? (const char *)menu->id->data : "", index) < 0)
         return FALSE;
-    /* FIXME: should pass show_empty into file somehow - v.1.2 may be */
+    /* pass show_empty into file if format is v.1.2 */
+    if (req_version >= 2 &&
+        fprintf(f, "%d\n", menu->layout.nodisplay ? FLAG_IS_NODISPLAY : 0) < 0)
+        return FALSE;
     for (child = menu->children; ok && child != NULL; child = child->next)
     {
         index = ((MenuApp *)child->data)->type;
@@ -929,6 +1044,16 @@ gboolean save_menu_cache(MenuMenu *layout, const char *menuname, const char *fil
     int i;
     gboolean ok = FALSE;
 
+    tmp = (char *)g_getenv("CACHE_GEN_VERSION");
+    if (tmp && sscanf(tmp, "%d.%u", &i, &req_version) == 2)
+    {
+        if (i != VER_MAJOR) /* unsupported format requested */
+            return FALSE;
+    }
+    if (req_version < VER_MINOR_SUPPORTED) /* unsupported format requested */
+        return FALSE;
+    if (req_version > VER_MINOR) /* fallback to maximal supported format */
+        req_version = VER_MINOR;
     all_apps = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, menu_app_free);
     for (i = 0; i < N_KNOWN_DESKTOPS; i++)
         DEs = g_slist_append(DEs, (gpointer)g_intern_static_string(de_names[i]));
@@ -953,7 +1078,7 @@ gboolean save_menu_cache(MenuMenu *layout, const char *menuname, const char *fil
     if (f == NULL)
         goto failed;
     /* Write common data */
-    fprintf(f, "1.1\n%s%s\n%d\n", /* FIXME: use CACHE_GEN_VERSION */
+    fprintf(f, "1.%d\n%s%s\n%d\n", req_version, /* use CACHE_GEN_VERSION */
             menuname, with_hidden ? "+hidden" : "",
             g_slist_length(DirDirs) + g_slist_length(AppDirs)
             + g_slist_length(MenuDirs) + g_slist_length(MenuFiles));
